/*********************************************************************
*                                                                    *
*  Shared Memory Definitions and Structures                          *
*                                                                    *
*********************************************************************/

#ifndef SHARED_MEM_H
#define SHARED_MEM_H

#include <object.h>

/* 
 * Shared Memory Object Subdirectory and Filename Suffix
 *
 * A subdirectory of OBJECT_DIR 
 */
#define SHARED_DIR                 "shared_mem/"
#define SHARED_MEM_FILENAME_SUFFIX ".shared"


/* 
 * Memory Descriptor Structure
 *
 * Used by create and delete top level routines 
 */
typedef struct memory_descriptor
{
  unsigned long  magic;
  int            type;
  unsigned long  address;
  unsigned long  length;
  int            reserved0;
  int            segment;
  char           name[4];
} MEM_DESC;

/* 
 * Magic number for Memory Descriptor
 */
#define MEM_MAGIC   0x87654321


/* 
 * Shared Memory Object Header 
 */
typedef struct memory_header
{
  int use_count;              /* total use count      */
  int memory_type;            /* nX or pSOS           */
  int pnn;                    /* PNN                  */
  int offset;                 /* Memory offset        */
  int length;                 /* Length in bytes      */
  unsigned long flags;        /* Mapping options      */
  int creator_cluster;
  int creator_id;
  int times_used;
  int times_released;
  int high_water;
} MEMORY_HEADER;


/* 
 * Shared Memory User Structure
 *
 * cluster ==  0:  Entry not in use
 * cluster == -1:  nX entry, ID = PID
 */
typedef struct memory_user
{
  int cluster;
  int id;
  int use_count;
} MEMORY_USER;

/* 
 * Maximum Number of Users per Shared Memory Object
 */
#define MAX_MEMORY_USERS    256

/* 
 * Shared Memory Object Structure
 */
typedef struct memory_object
{
  MEMORY_HEADER
    memory_header;
  MEMORY_USER
    memory_users[MAX_MEMORY_USERS];
} MEMORY_OBJECT;

/* 
 * Memory Types
 */
#define PSOS_MEMORY             	1
#define NX_MEMORY                 	2

/*
 * Shared Memory Mapping Options
 */
#define SM_CACHE_ENABLE                 0x00
#define SM_CACHE_INHIBIT                0x01

#define SM_CACHE_COPYBACK               0x00
#define SM_CACHE_WRITETHRU              0x02

#define SM_READ_WRITE                   0x00
#define SM_READ_ONLY                    0x04

#define SM_USER_ACCESS                  0x00
#define SM_SUPERVISOR_ONLY              0x08

#define SM_ANYWHERE                     0x00
#define SM_OBEY_VADDR                   0x10

/* 
 * Object Flag Definitions
 */
#define DELETE_ON_LAST_RELEASE  	0x01
#define DELETE_BY_CREATOR_ONLY  	0x02

/* 
 * Release Memory Object Options
 */
#define RPMO_RELEASE_ALL        	0x01
#define RPMO_DELETE_ON_LAST_RELEASE    	0x02	    
#define RPMO_RELEASE_DEAD_NX           	0x04
#define RPMO_RELEASE_DEAD_MACH          RPMO_RELEASE_DEAD_NX
#define RPMO_RELEASE_DEAD_CLUSTER       0x08
#define RPMO_RELEASE_UNCONDITIONAL      0x10

/* 
 * Delete Memory Object Options 
 */
#define DPMO_DELETE_BY_ANYONE           0x01
#define DPMO_ALREADY_LOCKED             0x02

#endif SHARED_MEM_H
