#ifndef _SIMLOCK_H
#define _SIMLOCK_H
/*****************************************************************************
File:     simlock.h
Contents: This file contains header information for a simple lock class
System:   BBNACI Programming Environments.
Created:  24-Nov-1989
Author:   ksedgwic

Remarks:

Copyright 1989 by BBN Advanced Computers Inc.
ALL RIGHTS RESERVED

No part of this material may be reproduced, stored in a retrieval
system, transmitted or used in any form or by any means, electronic,
mechanical, photocopying, recording, or otherwise, without the prior
written permission of, or express license from BBN Advanced Computers Inc.

RESTRICTED RIGHTS LEGEND

Use, duplication, or disclosure by the Government is subject to
restrictions as set forth in subparagraph (c)(1)(ii) of the Rights to
Technical Data and Computer Software clause at DFARS 52.227-7013.

BBN Advanced Computers Inc.
10 Fawcett Street
Cambridge, Ma. 02238
617-873-6000
*****************************************************************************/

/* $Header: simlock.h,v 1.1 90/01/15 11:54:54 ksedgwic Exp $ */

/****************************************************************
Class Simlock
****************************************************************/

#if defined(c_plusplus)

class simlock
{
    friend class sleeplock;	/* sleeplocks need to get at the fields */
    friend void sim_init(simlock *ptr);	/* c-interface intializer */
    
 private:
    int    interlock;		/* the interlock word */
    short  owner_nid;		/* the node id of the owner */
    short  owner_pid;		/* the process id of the owner */

    void retry();		/* try again to get the lock */
    void setup_process();	/* get our nid and pid cached */

 public:
           simlock(int skip = 0); /* the constructor */
    inline void lock();		/* lock the lock */
    inline void unlock();	/* unlock the lock */
};

#else /* defined(c_plusplus) */

struct simlock
{
    int   interlock;		/* the interlock word */
    short owner_nid;		/* the system logical node of the owner */
    short owner_pid;		/* the process id of the owner */
};
typedef struct simlock simlock;

#endif /* defined(c_plusplus) */

#ifndef NULL
#define NULL 0
#endif  NULL

#define NOINIT  1

/****************************************************************
Macros
****************************************************************/

#if defined(c_plusplus) || defined(__STDC__)
# ifdef BFLY1
extern  int   atomior32(int* ap, int v);
#define atomxmem(ap, v)		(atomior32((ap),(v)))
# endif BFLY1
# ifdef BFLY2
extern int    xmemi(int* ap, int v);
#define atomxmem(ap, v)		(xmemi((ap),(v)))
# endif BFLY2
#else /* defined(c_plusplus) || defined(__STDC__) */
# ifdef BFLY1
extern  int   atomior32();
#define atomxmem(ap, v)		(atomior32((ap),(v)))
# endif BFLY1
# ifdef BFLY2
extern int    xmemi();
#define atomxmem(ap, v)		(xmemi((ap),(v)))
# endif BFLY2
#endif /* defined(c_plusplus) || defined(__STDC__) */


/****************************************************************
Globals
****************************************************************/

extern int  sim_pid;		/* a cached copy of my pid */
extern int  sim_nid;		/* a cached copy of by system node id */
extern int  sim_pflag;		/* preemption control flag */
extern int  sim_ptime;		/* preemption control timer */


/****************************************************************
Event logging
****************************************************************/

#include <klog.h>
#define SIM_LEVEL  T3
#define logBase            10
#define simlockLockBegin   (logBase + 1)
#define simlockLockRetry   (logBase + 2)
#define simlockLockEnd     (logBase + 3)
#define simlockUnlockBegin (logBase + 4)
#define simlockUnlockEnd   (logBase + 5)


#if defined(c_plusplus)

/****************************************************************
simlock::lock(): acquire the lock
****************************************************************/

#define SIMLOCK_NPTIMEOUT (1000000) /* 1 second in usec */

inline void simlock::lock()
{
    register int oldval;

    if (!sim_pid)
	setup_process();

    KLOG_LOG(SIM_LEVEL, simlockLockBegin, (int) &interlock);

    /* Turn off preemption for a spell ... */
    if (sim_pflag == 0)
	sim_ptime = SIMLOCK_NPTIMEOUT;
    sim_pflag++;
    
    /* First, a quick try with no overhead */
    oldval = atomxmem(&interlock, 1);
    if (oldval == 0)
    {
	/* got it */
	owner_nid = sim_nid;
	owner_pid = sim_pid;
	KLOG_LOG(SIM_LEVEL, simlockLockEnd, (int) &interlock);
	return;
    }

    /* We didn't get it so we'll call the retry routine instead */
    retry();

}   /* end simlock::lock */
    

/****************************************************************
simlock::unlock(): unlock a simple lock.
****************************************************************/

inline void simlock::unlock()
{
    KLOG_LOG(SIM_LEVEL, simlockUnlockBegin, (int) &interlock);
    owner_nid = 0;
    owner_pid = 0;
    interlock = 0;
    sim_pflag--;
    KLOG_LOG(SIM_LEVEL, simlockUnlockEnd, (int) &interlock);

}   /* end simlock::unlock */

#endif /* defined(c_plusplus) */


/****************************************************************
C-Interface Routines
****************************************************************/

#if defined(c_plusplus) || defined(__STDC__)

void sim_init(simlock *lockptr);
void sim_lock(simlock *lockptr);
void sim_unlock(simlock *lockptr);

#else /* defined(c_plusplus) || defined(__STDC__) */

void sim_init();
void sim_lock();
void sim_unlock();

#endif /* defined(c_plusplus) || defined(__STDC__) */


/****************************************************************
sim_timeout: When a lock operation times out this routine is called.
If the routine returns, the lock operation will continue, starting
with a fresh timeout count.  A default sim_timeout is provided and 
may be overloaded with a user function as well.
****************************************************************/

#if defined(c_plusplus) || defined(__STDC__)

extern void sim_timeout(int mypid, int ownerpid, simlock *lptr);

#else /* defined(c_plusplus) || defined(__STDC__) */

extern void sim_timeout();

#endif /* defined(c_plusplus) || defined(__STDC__) */


/****************************************************************
util_atfork(): this routine is defined in simlock.cxx.  
IT IS IMPORTANT TO CALL THIS EITHER BEFORE OR AFTER YOU FORK!
If this routine is not called simplelocks will be labled incorrectly
and process will wait for locks in the incorrect manner.
****************************************************************/

extern void util_atfork();

/* ----- end of file marker ----- */
#endif _SIMLOCK_H
