#ifndef _SLEEPLOCK_H
#define _SLEEPLOCK_H
/*****************************************************************************
File:     sleeplock.h
Contents: This file contains header information for a sleepable lock class.
System:   BBNACI Programming Environments.
Created:  10-Dec-1989
Author:   ksedgwic

Remarks:

Copyright 1989 by BBN Advanced Computers Inc.
ALL RIGHTS RESERVED

No part of this material may be reproduced, stored in a retrieval
system, transmitted or used in any form or by any means, electronic,
mechanical, photocopying, recording, or otherwise, without the prior
written permission of, or express license from BBN Advanced Computers Inc.

RESTRICTED RIGHTS LEGEND

Use, duplication, or disclosure by the Government is subject to
restrictions as set forth in subparagraph (c)(1)(ii) of the Rights to
Technical Data and Computer Software clause at DFARS 52.227-7013.

BBN Advanced Computers Inc.
10 Fawcett Street
Cambridge, Ma. 02238
617-873-6000
*****************************************************************************/

/* $Header: sleeplock.h,v 1.1 90/01/15 11:55:51 ksedgwic Exp $ */

#include <stdio.h>
#include <simlock.h>


/****************************************************************
Sleeplock Class
****************************************************************/

#if defined(c_plusplus)
    
class sleeplock
{
    friend void slp_init(sleeplock *lockptr);
    
 private:
    simlock   thelock;
    short     state;
    short     owning_pid;

 public:
    sleeplock(int skipinit = 0); /* constructor */

    inline void lock();		/* lock the lock */
    inline void unlock();	/* unlock the lock */

    inline void latch();	/* latch the lock */
    inline void unlatch() { unlock(); }; /* unlatch the lock */

    inline void convert();	/* convert the latch to a lock */

};

#else /* defined(c_plusplus) */

struct sleeplock
{
    simlock   thelock;
    short     state;
    short     owning_pid;
};

typedef struct sleeplock sleeplock;

#endif /* defined(c_plusplus) */


/****************************************************************
Sleeplock States
****************************************************************/

#define SLEEPLOCK_FREE  0x0
#define SLEEPLOCK_BUSY  0x1
#define SLEEPLOCK_WAIT  0x2
#define SLEEPLOCK_LATCH 0x4


/****************************************************************
Prototypes
****************************************************************/

#if defined(c_plusplus)

extern int user_sleep_and_unlock(unsigned, long*, int);
extern void user_wakeup(unsigned);
extern void fprintf(FILE*, char* ...);
extern void perror(char*);
extern void exit(int);
extern void sigblock(int);

#endif /* defined(c_plusplus) */


/****************************************************************
Event Logging
****************************************************************/

#include <klog.h>
#define SLP_LEVEL  T3
#define sleeplockBase		20
#define sleeplockLockBegin   (sleeplockBase + 1)
#define sleeplockLockEnd     (sleeplockBase + 2)
#define sleeplockSleepBegin  (sleeplockBase + 3)
#define sleeplockSleepEnd    (sleeplockBase + 4)
#define sleeplockLatchBegin  (sleeplockBase + 5)
#define sleeplockLatchEnd    (sleeplockBase + 6)
#define sleeplockUnlockBegin (sleeplockBase + 7)
#define sleeplockUnlockEnd   (sleeplockBase + 8)
#define sleeplockConvert     (sleeplockBase + 9)
#define sleeplockWakeup      (sleeplockBase + 10)


#if defined(c_plusplus)

/****************************************************************
sleeplock::lock(): Lock the sleeplock.  If the lock is busy,
set the WAIT bit in the state and go to sleep on the sleeplock's
address as a wait channel.
****************************************************************/

inline void sleeplock::lock()
{
    int  usuret;

    KLOG_LOG(SLP_LEVEL, sleeplockLockBegin, (int) this);

    thelock.lock();

    while (state & SLEEPLOCK_BUSY)
    {
	KLOG_LOG(SLP_LEVEL, sleeplockSleepBegin, (int) this);

	state |= SLEEPLOCK_WAIT;
	usuret = user_sleep_and_unlock((unsigned) this,
				       (long*) &thelock.interlock, 0);
	switch (usuret)
	{
	case 3:			/* signal pending */
	    sigblock(0);	/* let os handle it */
	    /* intentionally fall through here */
	case 1:
	    KLOG_LOG(SLP_LEVEL, sleeplockSleepEnd, usuret);
	    thelock.lock();
	    break;

	case 2:			/* timeout? */
	case -1:		/* error */
	    fprintf(stderr, "sleeplock: usu unexepected return %d\n",
		    usuret);
	    perror("sleeplock: maybe its");
	    exit(1);
	}
    }
	    
    state = SLEEPLOCK_BUSY;
    owning_pid = thelock.owner_pid;

    thelock.unlock();
    KLOG_LOG(SLP_LEVEL, sleeplockLockEnd, (int) this);
    
}   /* end sleeplock::lock */
	

/****************************************************************
sleeplock::unlock():  Unlock a sleeplock.  If anyone is waiting
we'll issue a wakeup using the address of the sleeplock as the channel.
****************************************************************/

inline void sleeplock::unlock()
{
    KLOG_LOG(SLP_LEVEL, sleeplockUnlockBegin, (int) this);

    if ( !(state & SLEEPLOCK_LATCH) )
    {
	thelock.lock();

	if (state & SLEEPLOCK_WAIT)
	{
	    KLOG_LOG(SLP_LEVEL, sleeplockWakeup, (int) this);

	    user_wakeup((unsigned) this);
	}
    }

    state = SLEEPLOCK_FREE;
    owning_pid = 0;
    thelock.unlock();

    KLOG_LOG(SLP_LEVEL, sleeplockUnlockEnd, (int) this);

}   /* end sleeplock::unlock */



/****************************************************************
sleeplock::latch():  Acquire a sleep lock in latch mode.  We will
go to sleep if the state is BUSY.
****************************************************************/

inline void sleeplock::latch()
{
    int  usuret;

    KLOG_LOG(SLP_LEVEL, sleeplockLatchBegin, (int) this);

    thelock.lock();

    while (state & SLEEPLOCK_BUSY)
    {
	KLOG_LOG(SLP_LEVEL, sleeplockSleepBegin, (int) this);

	state |= SLEEPLOCK_WAIT;
	usuret = user_sleep_and_unlock((unsigned) this,
				       (long*) &thelock.interlock, 0);
	switch (usuret)
	{
	case 3:			/* signal pending */
	    sigblock(0);	/* let os handle it */
	    /* intentionally fall through here */
	case 1:

	    KLOG_LOG(SLP_LEVEL, sleeplockSleepEnd, usuret);
	    thelock.lock();
	    break;

	case 2:			/* timeout? */
	case -1:		/* error */
	    fprintf(stderr, "sleeplock: usu unexepected return %d\n",
		    usuret);
	    perror("sleeplock: maybe its");
	    exit(1);
	}
    }

    state = SLEEPLOCK_LATCH + SLEEPLOCK_BUSY;
    owning_pid = thelock.owner_pid;
    KLOG_LOG(SLP_LEVEL, sleeplockLatchEnd, (int) this);
	    
}   /* end sleeplock::latch */


/****************************************************************
sleeplock::convert():  Convert a sleeplock from latch mode to
lock mode.  Any processes currently spinning on the lock will
go to sleep.  
****************************************************************/

inline void sleeplock::convert()
{
    KLOG_TRIGGER();
    KLOG_LOG(SLP_LEVEL, sleeplockConvert, (int) this);

    state = SLEEPLOCK_BUSY;
    owning_pid = 0;
    thelock.unlock();

}   /* end sleeplock::convert */

#endif /* defined(c_plusplus) */


/****************************************************************
C-Interface routines: These routines implement a C subroutine
interface to the sleeplock methods.
****************************************************************/

#if defined(c_plusplus) || defined(__STDC__)

void slp_init(sleeplock *lockptr);
void slp_latch(sleeplock *lockptr);
void slp_unlatch(sleeplock *lockptr);
void slp_lock(sleeplock *lockptr);
void slp_unlock(sleeplock *lockptr);
void slp_convert(sleeplock *lockptr);

#else /* defined(c_plusplus) || defined(__STDC__) */

void slp_init();
void slp_latch();
void slp_unlatch();
void slp_lock();
void slp_unlock();
void slp_convert();

#endif /* defined(c_plusplus) || defined(__STDC__) */

/* ----- end of file marker ----- */
#endif _SLEEPLOCK_H
