#ifndef _UNSIGNED64_H
#define _UNSIGNED64_H
/*****************************************************************************
File:     unsigned64.h
Contents: This header file contains definitions for a 64 bit unsigned type.
System:   BBNACI Programming Environments.
Created:  4-Jan-1990
Author:   ksedgwic

Remarks: This file is a header file for both C and C++.  When used by
the C compiler it contains header information for an implmentation of
unsigned64 which is based on tricking the compiler into thinking they
are doubles.  When the file is used by C++ an actual unsigned64 class
is defined.

Copyright 1990 by BBN Advanced Computers Inc.
ALL RIGHTS RESERVED

No part of this material may be reproduced, stored in a retrieval
system, transmitted or used in any form or by any means, electronic,
mechanical, photocopying, recording, or otherwise, without the prior
written permission of, or express license from BBN Advanced Computers Inc.

RESTRICTED RIGHTS LEGEND

Use, duplication, or disclosure by the Government is subject to
restrictions as set forth in subparagraph (c)(1)(ii) of the Rights to
Technical Data and Computer Software clause at DFARS 52.227-7013.

BBN Advanced Computers Inc.
10 Fawcett Street
Cambridge, Ma. 02238
617-873-6000
*****************************************************************************/

/* $Header: unsigned64.h,v 1.1 90/01/15 11:56:12 ksedgwic Exp $ */

#if defined(c_plusplus)

/****************************************************************
C++ Header Information:  This section is used only when compiling 
with the C++ compiler.
****************************************************************/

/* ---------------- Class Definition ---------------- */

typedef double u64;

class unsigned64
{
 private:
    /* The values */
    unsigned int hi;
    unsigned int lo;

 public:
    /* Constructors */
    inline unsigned64(unsigned hival, unsigned loval);
    inline unsigned64(unsigned anunsigned);
    inline unsigned64(int anint);
    inline unsigned64();

    /* Predicates */
    inline int isint();

    /* Type Conversions */
    inline int operator int();
    inline unsigned operator unsigned();

    /* Comparison Operators */
    friend inline int operator==(unsigned64 arg1, unsigned64 arg2);
    friend inline int operator!=(unsigned64 arg1, unsigned64 arg2);
    friend inline int operator>(unsigned64 arg1, unsigned64 arg2);
    friend inline int operator>=(unsigned64 arg1, unsigned64 arg2);
    friend inline int operator<(unsigned64 arg1, unsigned64 arg2);
    friend inline int operator<=(unsigned64 arg1, unsigned64 arg2);
    
    /* Arithmetic Operators */
    friend inline unsigned64 operator+(unsigned64 arg1, unsigned64 arg2);
    friend inline unsigned64 operator-(unsigned64 arg1, unsigned64 arg2);
    friend inline unsigned64 operator*(unsigned64 arg1, unsigned64 arg2);

    /* Shift Operators */
    inline unsigned64 operator<<(int arg2);
    inline unsigned64 operator>>(int arg2);
    
};


/* ---------------- Constructors ---------------- */

inline unsigned64::unsigned64(unsigned hival, unsigned loval)
{
    hi = hival;
    lo = loval;
}

inline unsigned64::unsigned64(unsigned anunsigned)
{
    hi = 0;
    lo = anunsigned;
}

inline unsigned64::unsigned64(int anint)
{
    hi = 0;
    lo = (unsigned) anint;
}

inline unsigned64::unsigned64()
{
    lo = hi = 0;
}


/* ---------------- Predicates ---------------- */

inline int unsigned64::isint()
{
    return (! hi);
}


/* ---------------- Conversions ---------------- */ 

inline int unsigned64::operator int()
{
    return (int) this->lo;
}

inline unsigned unsigned64::operator unsigned()
{
    return this->lo;
}


/* ---------------- Comparisons ---------------- */

inline int operator==(unsigned64 arg1, unsigned64 arg2)
{
    return (arg1.lo == arg2.lo &&
	    arg1.hi == arg2.hi);
}

inline int operator!=(unsigned64 arg1, unsigned64 arg2)
{
    return (arg1.lo != arg2.lo ||
	    arg1.hi != arg2.hi);
}

inline int operator>(unsigned64 arg1, unsigned64 arg2)
{
    return ((arg1.hi > arg2.hi) ||
	    ((arg1.hi == arg2.hi) && (arg1.lo > arg2.lo))) ? 1 : 0;
}

inline int operator>=(unsigned64 arg1, unsigned64 arg2)
{
    return ((arg1.hi > arg2.hi) ||
	    ((arg1.hi == arg2.hi) && (arg1.lo >= arg2.lo))) ? 1 : 0;
}

inline int operator<(unsigned64 arg1, unsigned64 arg2)
{
    return ((arg1.hi < arg2.hi) ||
	    ((arg1.hi == arg2.hi) && (arg1.lo < arg2.lo))) ? 1 : 0;
}

inline int operator<=(unsigned64 arg1, unsigned64 arg2)
{
    return ((arg1.hi < arg2.hi) ||
	    ((arg1.hi == arg2.hi) && (arg1.lo <= arg2.lo))) ? 1 : 0;
}


/* ---------------- Arithmetic Operators ---------------- */

extern u64 u64add      (unsigned arg1hi, unsigned arg1lo,
			   unsigned arg2hi, unsigned arg2lo);
extern u64 u64subtract (unsigned arg1hi, unsigned arg1lo,
			   unsigned arg2hi, unsigned arg2lo);
extern u64 u64multiply (unsigned arg1hi, unsigned arg1lo,
			   unsigned arg2hi, unsigned arg2lo);

inline unsigned64 operator+(unsigned64 arg1, unsigned64 arg2)
{
    union {
	u64  as_u64;
	struct {
	    unsigned hi;
	    unsigned lo;
	} as_parts;
    } result;

    result.as_u64 = u64add(arg1.hi, arg1.lo, arg2.hi, arg2.lo);
    return unsigned64(result.as_parts.hi, result.as_parts.lo);

}

inline unsigned64 operator-(unsigned64 arg1, unsigned64 arg2)
{
    union {
	u64  as_u64;
	struct {
	    unsigned hi;
	    unsigned lo;
	} as_parts;
    } result;

    result.as_u64 = u64subtract(arg1.hi, arg1.lo, arg2.hi, arg2.lo);
    return unsigned64(result.as_parts.hi, result.as_parts.lo);

}

inline unsigned64 operator*(unsigned64 arg1, unsigned64 arg2)
{
    union {
	u64  as_u64;
	struct {
	    unsigned hi;
	    unsigned lo;
	} as_parts;
    } result;

    result.as_u64 = u64multiply(arg1.hi, arg1.lo, arg2.hi, arg2.lo);
    return unsigned64(result.as_parts.hi, result.as_parts.lo);

}

/* ---------------- Shift Operators ---------------- */

inline unsigned64 unsigned64::operator<<(int arg2)
{
    unsigned64 target;
    
    /* There are two different strategies depending on
       the magnitude of the shift.
       neg     => undefined
       0  - 31 => small shift
       32 - 63 => big shift
       64 -    => undefined
    */

    if (arg2 < 32)
    {
	/* This is the small shift case */

	register unsigned u;

	u = this->lo >> (32 - arg2); /* These are the bits that go to
				       the high word.*/
	target.lo = this->lo << arg2;
	target.hi = (this->hi << arg2) | u;
    }
    else
    {
	/* This is the big shift case */
	target.hi = this->lo << (arg2 - 32);
	target.lo = 0;
    }

    return target;
}

inline unsigned64 unsigned64::operator>>(int arg2)
{
    unsigned64 target;
    
    /* There are two different strategies depending on
       the magnitude of the shift.
       neg     => undefined
       0  - 31 => small shift
       32 - 63 => big shift
       64 -    => undefined
    */

    if (arg2 < 32)
    {
	/* This is the small shift case */

	register unsigned u;

	u = this->hi << (32 - arg2); /* These are the bits that go to
				       the low word.*/
	target.hi = this->hi >> arg2;
	target.lo = u | (this->lo >> arg2);
    }
    else
    {
	/* This is the big shift case */
	target.lo = this->hi >> (arg2 - 32);
	target.hi = 0;
    }

    return target;
}

#else /* defined(c_plusplus) */

/****************************************************************
C Header Information:  This section is used only when compiling 
with the C compiler.
****************************************************************/

typedef double u64;
typedef struct {
    unsigned  hi;
    unsigned  lo;
} unsigned64;

/* ---------------- Define Functions ---------------- */

extern u64 u64fromints(/* int hival, int loval */);
extern u64 u64fromint(/* int theint */);
extern int u64toint(/* u64 *arg */);
extern int u64isint(/* u64 *checkme */);
extern int u64isequal(/* u64 *a, u64 *b */);
extern int u64isnotequal(/* u64 *a, u64 *b */);
extern int u64isgreater(/* u64 *a, u64 *b */);
extern int u64isgreaterequal(/* u64 *a, u64 *b */);
extern int u64isless(/* u64 *a, u64 *b */);
extern int u64islessequal(/* u64 *a, u64 *b */);
extern u64 u64add(/* u64 *a, u64 *b */);
extern u64 u64subtract(/* u64 *a, u64 *b */);
extern u64 u64multiply(/* u64 *a, u64 *b */);
extern u64 u64shiftl(/* u64 *shiftme, int count */);
extern u64 u64shiftr(/* u64 *shiftme, int count */);

#endif /* defined(c_plusplus) */

/* ----- end of file marker ----- */
#endif _UNSIGNED64_H
