/* $Header: /nfs/europa/u1/BFMACH.local/mk/h/RCS/kmalloc.h,v 1.7 90/11/09 12:54:28 tappan Exp $ */
/*
 * Kernel memory allocator.
 * Data structures, constants, and macros.
 */

#ifndef _kmalloc_h
#define _kmalloc_h yes
#ifdef KERNEL
#include "../h/types.h"
#include "../h/lock.h"
/*
 * KMALLOC_DUPALLOCDEBUG determines whether code is included to
 * debug multiple allocations and frees.
 * This adds extra code to the KMALLOC and KFREE macros and
 * increases the size of the chunk map.
 *
 * define non-zero to include the code.
 */
/*
 * For now, we are overriding the feature setting
 */
#define KMALLOC_DUPALLOCDEBUG 1
#else KERNEL
#include <sys/types.h>
#include <sys/lock.h>
#endif KERNEL

#ifndef VECTORIZED
#define VECTORIZED extern
#endif VECTORIZED

/*
 * Memory allocation pools. 
 * These are currently used only for gathering usage information.
 */
/* The first set is derived from the old mbuf usage */
#define	KM_MT_FREE	0x00	/* should be on free list */
#define KM_MT_MBUF	0x01	/* From MBUF allocator */
#define KM_MT_MCL	0x02	/* cluster mbufs */
#define KM_MT_MCLREF	0x03	/* cluster mbuf reference counts */
#define KM_MT_NFS	0x04	/* NFS data structures */
#define KM_MT_CLUSTER	0x05	/* cluster() call related */
#define KM_MT_VFS	0x06	/* VFS related */
#define KM_MT_MISC      0x07	/* Miscellaneous */
#define KM_MT_BUF	0x08	/* buffer pool related */
#define KM_NMT		0x09	/* Number of pools */ 

/*
 * A allocation/free request can pass in a subtype field in the
 * high word of the type argument.
 */
#define KMALLOC_TYPE_MASK 0xffff
#define KMALLOC_SUBTYPE_SHIFT 16
#define KMALLOC_SUBTYPE_MASK 0xffff0000
#define KMALLOC_SUBTYPE(x) (((x)&KMALLOC_SUBTYPE_MASK)>>KMALLOC_SUBTYPE_SHIFT)
#define KMALLOC_TYPE(x) ((x)&KMALLOC_TYPE_MASK)
#define KMALLOC_MAKETYPE(t, s) ((t)|((s)<<KMALLOC_SUBTYPE_SHIFT))
/*
 * The kernel memory allocator uses a number of different 'buckets'.
 * All but the largest are fixed size.
 *
 * The KMALLOC_BUCKETOF() macro decodes which bucket to use for a given 
 * size request. Although it looks ugly, the compiler can optimize
 * this away for constant size requests.
 */

/*
 * Initial implementation - the following bucket sizes are used
 * >= 32
 * 33-64
 * 64-128
 * 128-256
 * 256-512
 * 512-1024
 * 1024-2048
 * > 2048
 */
#define KMALLOC_NBUCKETS 8
#define KMALLOC_BUCKETOF(n) ((n) <= 32 ? 0 : \
		     ((n) <= 64 ? 1 : \
		      ((n) <= 128 ? 2 : \
		       ((n) <= 256 ? 3 : \
			((n) <= 512 ? 4 : \
			 ((n) <= 1024 ? 5 : \
			  ((n) <= 2048 ? 6 : 7)))))))

#define KMALLOC_MAXBUCKET 7		/* Maximum bucket index */

/*
 * KMALLOC_MINBLOCK is the minimum block size which can be allocated
 * This affects the size of the duplicate allocation debugging
 * bitmap allocated in the chunk map (below)
 */
#define KMALLOC_MINBLOCK 32

/*
 * Memory is allocated in units of KMALLOC_CHUNKSIZE,
 * these can be further divided but all blocks within one
 * chunk are the same size, and belong to the same bucket.
 */
#define KMALLOC_CHUNKSIZE 512

/*
 * Memory is allocated out of blocks of a constant size
 * KMALLOC_EXPAND which are 'stolen' from the kernel virtual memory on
 * a node. 
 * These blocks must be aligned on a KMALLOC_EXPAND boundry so, given
 * any address in the block, the address of the block can be found.
 * KMALLOC_EXPAND must be greater than the size of the maximum allocation
 * by at least the size of the chunk map.
 */
#define KMALLOC_EXPAND (128*1024)

/*
 * KMALLOC_BLOCKTOBIT returns the bit number in the map's allocation
 * array associated with a given block.
 */
#define KMALLOC_BLOCKTOBIT(p) \
    (((u_long)(p)&(KMALLOC_EXPAND-1))/(KMALLOC_MINBLOCK))

/*
 * The first chunk is used to hold a map indicating which bucket a 
 * given block belongs to.
 * 
 * For fixed size buckets the first map entry points to the bucket to
 * free to.
 *
 * For blocks in MAXBUCKET (which must be >= 2*CHUNKSIZE in size)
 * the first map entry points to the bucket, the second holds the
 * actual block size.
 *
 */
typedef struct {
    u_short magic;		/* Magic number */
#define KMALLOC_MAGIC 0x1379
    u_short node;		/* Node owning map */
    struct kmalloc_bucket *bucket[KMALLOC_EXPAND/KMALLOC_CHUNKSIZE - 1];
    /* The following bitmap is used for debugging duplicate
     * allocations and frees. 
     * There is one bit for each minimally sized block.
     */
    /* Bitmap of free blocks */
    u_char allocatedblocks[(KMALLOC_EXPAND/(KMALLOC_MINBLOCK*8))+1];
} kmalloc_chunkmap;


/*
 * Given a block address, the address of its map can be found
 */
#define KMALLOC_BLOCKTOMAP(p) \
  (kmalloc_chunkmap *)((u_long)(p) & ~(KMALLOC_EXPAND-1))

/*
 * And also the starting chunk number
 * Chunks are offset by 1 because the first chunk of an allocation
 * block always holds the chunk map.
 */
#define KMALLOC_BLOCKTOCHUNK(p) \
  ((((u_long)(p)&(KMALLOC_EXPAND-1))/KMALLOC_CHUNKSIZE)-1)

/*
 * Conceptually items on a bucket's free list have the following
 * structure. In anything but the MAXBUCKET only the link field
 * is used.
 *
 * Note that if debugging code is enabled this structure implies
 * the minimum block size is 16 bytes.
 */
typedef struct kmalloc_bucket_item {
    struct kmalloc_bucket_item *link;
    u_int size;			/* size of this item */
    u_long freer;
    u_long freer2;
} kmalloc_bucket_item, *kmalloc_bucket_item_ptr;

/*
 * Structure of a memory allocation bucket
 */
typedef struct kmalloc_bucket {
    u_short bucket_size;	/* size of items in this bucket */
    u_short bucket_expandlimit;	/* limit on the number of times this bucket */
    u_int bucket_allocsize;	/* allocation size for items (>= bucket_size)*/
    /* can be expanded. expand_limit*allocsize */
    /* defines the maximum storage a bucket can use */
    u_short bucket_node;	/* Node this bucket resides on */
    u_char bucket_expandbucket, /* Bucket that expansions come from */
           bucket_bucket,	/* bucket number of this bucket */
          /* Flags */
           bucket_waiting, /* A process is waiting for storage */
           bucket_expanding,	/* A process is expanding bucket's allocation*/
           bucket_needexpand,	/* bucket needs to be expanded */
           bucket_maxbucket,	/* This is MAXBUCKET */
           bucket_spare1, bucket_spare2;
    int bucket_reserved;	/* reserved space */
    simple_lock_data_t bucket_lock; /* multiprocessor lock for this bucket */
    kmalloc_bucket_item *bucket_freelist; /* free block list */
    u_int bucket_use[KM_NMT];	/* Memory usage statistics for this pool */
} kmalloc_bucket;

#ifdef KERNEL
/*
 * memory allocations must be done at elevated priority, since they
 * can occur at interrupt level (cf. network mbufs)
 */
#define spl_kmalloc() splhigh()
/*
 * macros for allocating and freeing memory
 * KMALLOC - allocates a block of memory on the current node,
 * cast to the appropriate type.
 * KMALLOC_ON_NODE - allocates a block of memory on an arbitrary node.
 * KFREE - frees a block of memory.
 */
VECTORIZED int my_cpu_number;
VECTORIZED kmalloc_bucket *kmalloc_buckets;

/*
 * Possible values for the 'canwait' field
 */
#define KMALLOC_WAIT 1
#define KMALLOC_NOWAIT 0

#define KMALLOC_ON_NODE(p, cast, size, canwait, node, type) \
    p = (cast)_kmalloc(KMALLOC_BUCKETOF(size), size, canwait, node, type)
				      
#define KMALLOC(p, cast, size, canwait, type) \
  KMALLOC_ON_NODE(p, cast, size, canwait, my_cpu_number, type)

#define KFREE(p, type) \
    _kfree(p, type, KMALLOC_BLOCKTOMAP((p)), KMALLOC_BLOCKTOCHUNK((p)))

#define KMALLOC_ON_NODE_ZEROED(p, cast, size, canwait, node, type) \
{ \
      KMALLOC_ON_NODE(p, cast, size, canwait, node, type); \
      bzero((p), (size)); \
}

#define KMALLOC_ZEROED(p, cast, size, canwait, type) \
    KMALLOC_ON_NODE_ZEROED(p, cast, size, canwait, my_cpu_number, type)

/*
 * KMALLOC_AVAILABLE - determines whether there currently free space in
 * the bucket which holds requests of the indicated size.
 *
 * Caution. This is not interlocked. There is no guarantee that the
 * space will actually be there when you try to acquire it.
 */
#define KMALLOC_AVAILABLE(size) \
    _kmalloc_available(KMALLOC_BUCKETOF(size), my_cpu_number)

#define KMALLOC_AVAILABLE_ON_NODE(size, lnode) \
    _kmalloc_available(KMALLOC_BUCKETOF(size), lnodeb)
    
#endif KERNEL

#ifdef KERNEL
void kfree();
kmalloc_chunkmap *kmalloc_blocktomap();
int kmalloc_blocktochunk();
kmalloc_bucket *kmalloc_bucketof();
kmalloc_bucket_item *_kmalloc();
kmalloc_bucket_item *kmalloc(), *kmalloc_on_node();
kmalloc_bucket_item *kmalloc_zeroed(), *kmalloc_zeroed_on_node();
void _kfree();
int _kmalloc_available();

/*
 * The following are gistpoints, used when the KMALLOC_GISTLEVEL
 * feature is enabled.
 */
#define ELOG_KMALLOC_ALLOC 100          /* size                         %d */
#define ELOG_KMALLOC_FREE  101          /* size                         %d */
#define ELOG_KMALLOC_EXPAND 102         /* bucket                       %d */
#define ELOG_KMALLOC_MAXBUCKETALLOC 103 /* size                         %d */
#define ELOG_KMALLOC_DENIED 104         /* size                         %d */
#define ELOG_KMALLOC_ALLOCTYPE 105	/* type                         %x */
#define ELOG_KMALLOC_FREETYPE 106       /* type                         %x */

#endif KERNEL

/*
 * Emacs C format control
 *
 * Local Variables:
 * c-indent-level: 4
 * c-continued-statement-offset: 4
 * c-brace-offset: -4
 * c-argdecl-indent: 5
 * c-label-offset: -4
 * End:
 */
#endif _kmalloc_h
